from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Type, TYPE_CHECKING
from depydency.inject import Inject

if TYPE_CHECKING:
    from depydency.abc_container import AbcContainer


class AbcProvider(ABC):
    _container: AbcContainer

    @abstractmethod
    def provide(self, inject: Inject) -> Any:
        pass

    @abstractmethod
    def get_dependency_type(self) -> Type:
        pass

    def get_container(self) -> AbcContainer:
        return self._container

    def set_container(self, container: AbcContainer):
        self._container = container



from typing import Type


class AbcInstanceException(Exception):

    def __init__(self, dependency_type: Type, previous: Exception):
        self.dependency_type = dependency_type
        self.previous = previous

        super().__init__(
            f"Can't create instance of {dependency_type}. "
            f"If it's abstract, provide an implementation via the "
            f"Alias provider or set a default implementation via "
            f"TypeInject. Reason: {previous}"
        )



from typing import Type


class BadTypeException(Exception):

    def __init__(self, type_expected: Type, type_provided: Type):
        self.type_expected = type_expected
        self.type_provided = type_provided

        super().__init__(
            f"Expected class {type_expected}, got {type_provided} from provider"
        )



from typing import Any, Type, TYPE_CHECKING
from depydency.provider.abc_provider import AbcProvider
from depydency.inject import Inject


class Value(AbcProvider):
    def __init__(self, value: Any):
        self._value = value

    def provide(self, inject: Inject) -> Any:
        assert inject.unique_instance == False, "Can only inject the actual value"
        return self._value

    def get_dependency_type(self) -> Type:
        return type(self._value)



from typing import Any, Type
from depydency.provider.abc_provider import AbcProvider
from depydency.inject import Inject


class Alias(AbcProvider):
    _alias_type: Type
    _target_type: Type

    def __init__(self, alias_type: Type, target_type: Type):
        self._alias_type = alias_type
        self._target_type = target_type

    def provide(self, inject: Inject) -> Any:
        inject.set_dependency_id(dependency_type=self._target_type)
        return self._container.get_dependency(inject)

    def get_dependency_type(self) -> Type:
        return self._alias_type



from typing import Type, Any
from depydency.abc_locator_interface import AbcLocatorInterface
from depydency.inject import Inject


class StaticLocator:

    locator: AbcLocatorInterface

    @classmethod
    def get_by_type[DependencyType](
        cls, dependency_type: Type[DependencyType],
        unique_instance: bool = False,
        default_implementation: Type | None = None,
    ) -> DependencyType:
        return cls.locator.get_by_type(dependency_type, unique_instance, default_implementation)
        
    @classmethod
    def get_by_name(
        cls, dependency_name: str,
        unique_instance: bool = False,
        default_value: Any = None,
    ) -> Any:
        return cls.locator.get_by_name(dependency_name, unique_instance, default_value)



class BadNameException(Exception):

    def __init__(self, dependency_name: str):
        self.dependency_name = dependency_name
        super().__init__(
            f"Named dependency '{dependency_name}' does not exist"
        )



from depydency.inject import Inject
from depydency.provider.abc_creator import AbcCreator
from typing import Type, Any, Callable

CreatorCallable = Callable[[AbcCreator, Inject], Any]


class Callback(AbcCreator):
    def __init__(self, dependency_type: Type, creator_callable: CreatorCallable):
        assert callable(creator_callable), "Pass a callable"

        self._dependency_type = dependency_type
        self._creator_callable = creator_callable
        self._instance = None

    def _creator(self, inject: Inject) -> Any:
        return self._creator_callable(self, inject)



from typing import Dict
from depydency.inject import Inject
from depydency.exceptions.bad_type import BadTypeException
from depydency.provider.abc_provider import AbcProvider
from abc import ABC, abstractmethod

from typing import Any, Type, Annotated, get_type_hints, get_origin, get_args


class AbcCreator(AbcProvider, ABC):
    _dependency_type: Type
    _instance: Any

    def get_dependency_type(self) -> Type:
        return self._dependency_type

    def provide(self, inject: Inject) -> Any:
        if inject.unique_instance:
            return self._create_instance(inject)
        
        if self._instance is None:
            self._instance = self._create_instance(inject)

        return self._instance

    def inject_dependencies(self, instance: Any, dependencies: Dict[str, Any] = {}):
        instance_type = type(instance)
        annotations = get_type_hints(instance_type, include_extras=True)
        
        for dependency_name, annotation in annotations.items():
            if get_origin(annotation) == Annotated:
                dependency_type, inject = get_args(annotation)
                if (isinstance(dependency_type, Type) and isinstance(inject, Inject)):
                    inject.set_dependency_id(dependency_type, dependency_name)
                    dependency_instance = dependencies.get(dependency_name)
                    
                    if not dependency_instance:
                        dependency_instance = self._container.get_dependency(inject)
                    elif not isinstance(dependency_instance, inject.dependency_type):
                        raise BadTypeException(
                            inject.dependency_type, type(dependency_instance)
                        )
                    setattr(instance, dependency_name, dependency_instance)

    def _create_instance(self, inject: Inject) -> Any:
        instance = self._creator(inject)
        type_provided = type(instance)
        if type_provided != self._dependency_type:
            raise BadTypeException(self._dependency_type, type_provided)

        return instance

    @abstractmethod
    def _creator(self, inject: Inject) -> Any:
        pass



from abc import ABC, abstractmethod
from typing import Dict, Type, Any
from depydency.abc_locator_interface import AbcLocatorInterface
from depydency.provider.abc_provider import AbcProvider
from depydency.provider.auto_resolve import AutoResolve
from depydency.exceptions.bad_name import BadNameException
from depydency.exceptions.abc_instance import AbcInstanceException
from depydency.inject import Inject, TypeInject, NameInject


class AbcContainer(AbcLocatorInterface, ABC):

    _providers_type: Dict[str, AbcProvider]
    _providers_name: Dict[str, AbcProvider]

    def __init__(self):
        self._providers_type = {}
        self._providers_name = {}
        self.setup()

    @abstractmethod
    def setup(self):
        """Setup dependences in implementation"""

    def provide_type(self, provider: AbcProvider):
        type_repr = self._get_type_repr(provider.get_dependency_type())
        self._providers_type[type_repr] = provider
        provider.set_container(self)

    def provide_name(self, name: str, provider: AbcProvider):
        self._providers_name[name] = provider

    def get_by_type[DependencyType](
        self, dependency_type: Type[DependencyType],
        unique_instance: bool = False,
        default_implementation: Type | None = None,
    ) -> DependencyType:
        inject = TypeInject(unique_instance, default_implementation)
        inject.set_dependency_id(dependency_type=dependency_type)
        return self.get_dependency(inject)
        
    def get_by_name(
        self, dependency_name: str,
        unique_instance: bool = False,
        default_value: Any = None
    ) -> Any:
        inject = NameInject(unique_instance, default_value)
        NameInject.set_dependency_id(dependency_name=dependency_name)
        return self.get_dependency(inject)

    def get_dependency(self, inject: Inject) -> Any:
        match inject.method:
            case Inject.Method.BY_TYPE:
                type_repr = self._get_type_repr(inject.dependency_type)
                if type_repr not in self._providers_type.keys():
                    self.provide_type(AutoResolve(inject.dependency_type))
                try:
                    return self._providers_type.get(type_repr).provide(inject)
                except TypeError as exception:
                    if inject.default_implementation is None:
                        raise AbcInstanceException(inject.dependency_type, exception)
                    default_repr = self._get_type_repr(inject.default_implementation)
                    if default_repr not in self._providers_type.keys():
                        self.provide_type(AutoResolve(inject.default_implementation))
                    return self._providers_type.get(default_repr).provide(inject)
            case Inject.Method.BY_NAME:
                if inject.dependency_name not in self._providers_name.keys():
                    if inject.default_value is not None:
                        return inject.default_value
                    raise BadNameException(inject.dependency_name)
                return self._providers_name.get(inject.dependency_name).provide(inject)
            
        raise ValueError(f"Unexpected inject method {inject.method}")

    @staticmethod
    def _get_type_repr(dependency_type: Type) -> str:
        return f"{dependency_type.__module__}.{dependency_type.__name__}"



from abc import ABC, abstractmethod
from typing import Type, Any


class AbcLocatorInterface(ABC):

    @abstractmethod
    def get_by_type[DependencyType](
        self, dependency_type: Type[DependencyType], 
        unique_instance: bool = False,
        default_implementation: Type | None = None,
    ) -> DependencyType:
        pass
        
    @abstractmethod
    def get_by_name(
        self, dependency_name: str, 
        unique_instance: bool = False,
        default_value: Any = None,
    ) -> Any:
        pass



from abc import ABC
from enum import Enum
from typing import Type, Any


class Inject(ABC):
    class Method(str, Enum):
        BY_TYPE: str = 'by_type'
        BY_NAME: str = 'by_name'

    method: Method
    unique_instance: bool

    # Method.BY_TYPE:
    dependency_type: Type
    default_implementation: Type | None

    # Method.BY_NAME:
    dependency_name: str | None
    default_value: Any

    def __init__(self):
        raise Exception("Use TypeInject or NameInject")
    
    def set_dependency_id(
        self, 
        dependency_type: Type = Any, 
        dependency_name: str | None = None,
    ):
        self.dependency_type = dependency_type
        self.dependency_name = dependency_name


class TypeInject(Inject):
    def __init__(
        self,
        unique_instance: bool = False,
        default_implementation: Type | None = None,
    ):
        self.method = self.Method.BY_TYPE
        self.unique_instance = unique_instance
        self.default_implementation = default_implementation


class NameInject(Inject):
    def __init__(
        self,
        unique_instance: bool = False,
        default_value: Any = None,
    ):
        self.method = self.Method.BY_NAME
        self.unique_instance = unique_instance
        self.default_value = default_value



from typing import Type, Any
from depydency.provider.abc_creator import AbcCreator
from depydency.inject import Inject
import importlib


class AutoResolve(AbcCreator):
    def __init__(self, dependency_type: Type):
        self._dependency_type = dependency_type
        self._instance = None

    def _creator(self, inject: Inject) -> Any:
        module = importlib.import_module(self.get_dependency_type().__module__)
        type_reference = getattr(module, self.get_dependency_type().__name__)
        
        instance = type_reference()
        self.inject_dependencies(instance)
        
        return instance
